using GriffinNav.GPS;
using GriffinNav.UI;
using System;
using System.Text;
using System.Windows.Forms;
using Microsoft.Win32;

namespace GriffinNav.UI
{
    public partial class frmGpsData : Form
    {
        private bool mGpsReInit = false;

        public frmGpsData()
        {
            InitializeComponent();

            try
            {
                // check for existing port setting
                if (!string.IsNullOrEmpty(AppSettings.GpsPort))
                {
                    lblPort.Text = AppSettings.GpsPort;
                }

                Text = Global.Name;
            }
            catch (Exception ex)
            {
                Global.ShowError("Error", ex.Message);
            }
        }

        public DialogResult ShowDialog(out bool gpsReInit)
        {
            base.ShowDialog();

            gpsReInit = mGpsReInit;

            return DialogResult.None;
        }

        private void Gps_VelocityChanged(Velocity velocity)
        {
            if (InvokeRequired)
            {
                try
                {
                    VelocityChangedHandler d = new VelocityChangedHandler(Gps_VelocityChanged);
                    Invoke(d, new object[] { velocity });
                }
                catch (ObjectDisposedException)
                { }
            }
            else
            {
                // check which unit system to use
                if (AppSettings.UnitSystem == UnitSystem.English)
                {
                    // show speed in miles per hour
                    lblSpeed.Text = decimal.Round(velocity.MPH, 0).ToString() + " mph";
                }
                else
                {
                    // show speed in kilometersper hour
                    lblSpeed.Text = decimal.Round(velocity.KPH, 0).ToString() + " kph";
                }

                // show heading direction and angle
                lblHeading.Text = velocity.Angle.ToString() + " (" + velocity.Direction.ToString() + ")";
            }
        }

        private void Gps_PositionChanged(Position position)
        {
            if (InvokeRequired)
            {
                try
                {
                    PositionChangedHandler d = new PositionChangedHandler(Gps_PositionChanged);
                    Invoke(d, new object[] { position });
                }
                catch (ObjectDisposedException)
                { }
            }
            else
            {
                lblSatellites.Text = position.Satellites.ToString(); // number of satellites being tracks

                // latitude/longitude in degrees & minutes
                lblLatitude.Text = position.Latitude.Degrees.ToString() + " " + position.Latitude.Minutes.ToString() + "' " + position.Latitude.Direction.ToString();
                lblLongitude.Text = position.Longitude.Degrees.ToString() + " " + position.Longitude.Minutes.ToString() + "' " + position.Longitude.Direction.ToString();
                
                lblFix.Text = position.FixQualityString; // GPS fix quality

                //check which unit system to use
                if (AppSettings.UnitSystem == UnitSystem.English)
                {
                    // show altitude in feet
                    lblAltitude.Text = position.Altitude.Feet.ToString() + " ft";
                }
                else
                {
                    // show altitude in meters
                    lblAltitude.Text = position.Altitude.Meters.ToString() + " m";
                }
            }
        }

        private void mnuMain_StartStop_Click(object sender, EventArgs e)
        {
            bool startStop = true;

            try
            {
                if (string.IsNullOrEmpty(AppSettings.GpsPort))
                {
                    Global.ShowError("", "You must select a COM port before starting the GPS.");
                    return;
                }

                if (AppSettings.GpsBaudRate <= 0)
                {
                    Global.ShowError("", "You must select a baud rate before starting the GPS.");
                    return;
                }
                
                // check if GPS is initialized and initialize if it is not
                if (Global.Gps == null)
                {
                    startStop = Global.InitGps(false); 
                    subscribeToEvents();
                    mGpsReInit = true;
                }
                else
                {
                    startStop = true;
                }

                // check if allowed to continue
                if (startStop)
                {
                    try
                    {
                        // disable Start/Stop button
                        mnuMain_StartStop.Enabled = false;

                        // check GPS state
                        if (Global.Gps.GpsState == DeviceState.Running)
                        {
                            Global.Gps.StopGps(); // STOP
                        }
                        else
                        {
                            Global.Gps.StartGps(AppSettings.GpsPort, AppSettings.GpsBaudRate, AppSettings.GpsReadTimeout); // START
                        }
                    }
                    catch (Exception ex)
                    {
                        Global.Gps = null;

                        // reset interface
                        mnuMain_StartStop.Text = "Start";
                        mnuMain_StartStop.Enabled = true;

                        Global.ShowError("Error", ex.Message);
                    }
                }
            }
            catch (Exception ex)
            {
                Global.ShowError("Error", ex.Message);
            }
        }

        private void mnuMainSettings_Click(object sender, EventArgs e)
        {
            try
            {
                // show GPS settings
                Global.Options.ShowDialog(OptionsTab.GPS);

                // check for port setting
                if (!string.IsNullOrEmpty(AppSettings.GpsPort))
                {
                    lblPort.Text = AppSettings.GpsPort;
                }
            }
            catch (Exception ex)
            {
                Global.ShowError("Error", ex.Message);
            }
        }

        private void subscribeToEvents()
        {
            Global.Gps.GpsStarted += new GenericHandler(Gps_GpsStarted);
            Global.Gps.GpsStopped += new GenericHandler(Gps_GpsStopped);
            Global.Gps.PositionChanged += new PositionChangedHandler(Gps_PositionChanged);
            Global.Gps.VelocityChanged += new VelocityChangedHandler(Gps_VelocityChanged);
            Global.Gps.OnGpsError += new GpsErrorHandler(Global.Gps_OnGpsError);
        }

        private void Gps_GpsStarted()
        {
            if (InvokeRequired)
            {
                try
                {
                    GenericHandler d = new GenericHandler(Gps_GpsStarted);
                    Invoke(d);
                }
                catch (ObjectDisposedException) { }
            }
            else
            {
                // update interface
                mnuMain_StartStop.Enabled = true;
                mnuMain_StartStop.Text = "Stop";
            }
        }

        private void Gps_GpsStopped()
        {
            if (InvokeRequired)
            {
                try
                {
                    GenericHandler d = new GenericHandler(Gps_GpsStopped);
                    Invoke(d);
                }
                catch (ObjectDisposedException) { }
            }
            else
            {
                // clear non-position related GPS data from interface
                lblSatellites.Text = "-";
                lblFix.Text = "-";
                lblHeading.Text = "-";
                lblSpeed.Text = "-";

                mnuMain_StartStop.Enabled = true;
                mnuMain_StartStop.Text = "Start";
            }
        }
    }
}